import abc
import contextlib
import dataclasses
import logging
import os
import subprocess
import sys
from typing import (
    Callable,
    Generator,
    Generic,
    Iterable,
    MutableMapping,
    ParamSpec,
    Sequence,
    TypeVar,
)

logger = logging.getLogger('faterunner')
T = TypeVar('T')
P = ParamSpec('P')


@dataclasses.dataclass(kw_only=True)
class Opts:
    silent: bool | None = None
    ignore_err: bool | None = None
    keep_going: bool | None = None
    dry: bool | None = None

    def __or__(self, other: 'Opts | None') -> 'Opts':
        if other is None:
            return self.__class__(**self.__dict__)
        update = {k: v for k, v in other.__dict__.items() if v is not None}
        new_dict = self.__dict__ | update
        return self.__class__(**new_dict)


class Action(abc.ABC):
    opts: Opts

    @abc.abstractmethod
    def _run(self, opts: Opts) -> None: ...

    @abc.abstractmethod
    def get_name(self) -> str: ...

    def run(self, opts: Opts | None = None) -> None:
        opts = self.opts | opts

        logger.debug(f'Current action: {self}')
        logger.debug(f'Action options: {opts}')
        logger.info(self.get_name())
        if opts.dry:
            return

        if opts.silent:
            output_redirect = self.redirect_stdout_stderr()
        else:
            output_redirect = contextlib.nullcontext()
        try:
            # NOTE: if you want you can capture output in silent
            # instead of devnull'ing it and attach it to the err if raised
            with output_redirect:
                self._run(opts)

        except Exception as err:
            # because ignore_err is supposed to suppress any error in execution
            # we have to deal with it here
            # i.e. we can't let it go to Manager._run where all the logging is
            # so we have to log it here ourselves
            if not opts.ignore_err:
                raise err
            logger.info(f'{err} (ignored)')

    @contextlib.contextmanager
    def redirect_stdout_stderr(self) -> Generator[None, None, None]:
        with (
            open(os.devnull, 'w') as nullout,
            open(os.devnull, 'w') as nullerr,
        ):
            with contextlib.redirect_stdout(nullout):
                with contextlib.redirect_stderr(nullerr):
                    yield


class SubprocessAction(Action):
    def __init__(self, cmd: str, opts: Opts | None = None) -> None:
        if not opts:
            opts = Opts()

        self.cmd = cmd
        self.opts = opts

    def get_name(self) -> str:
        return self.cmd

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({repr(self.cmd)})'

    def _run(self, opts: Opts) -> None:
        _ = opts
        proc = subprocess.run(
            self.cmd,
            stdout=sys.stdout,
            stderr=sys.stderr,
            shell=True,
        )
        proc.check_returncode()


class FunctionAction(Action, Generic[P, T]):
    def __init__(
        self,
        func: Callable[P, T],
        opts: Opts | None = None,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> None:
        if not opts:
            opts = Opts()

        self.func = func
        self.opts = opts
        self.args = args
        self.kwargs = kwargs

    def get_name(self) -> str:
        return self.func.__name__

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({repr(self.func)})'

    def _run(self, opts: Opts) -> None:
        _ = opts
        self.func(*self.args, **self.kwargs)


class Task:
    def __init__(
        self, actions: Sequence[Action], opts: Opts | None = None
    ) -> None:
        if not opts:
            opts = Opts()

        self.actions = actions
        self.opts = opts

    def run(self, opts: Opts | None = None) -> None:
        opts = self.opts | opts
        logger.debug(f'Task options: {opts}')

        for action in self.actions:
            action.run(opts)


class Manager:
    def __init__(
        self,
        tasks: MutableMapping[str, Task] | None = None,
        opts: Opts | None = None,
    ) -> None:
        if not tasks:
            tasks = {}
        if not opts:
            opts = Opts()

        self.tasks = tasks
        self.deps: dict[str, Iterable[str]] = {}
        self.opts = opts

    def add(
        self, name: str, task: Task, deps: Iterable[str] | None = None
    ) -> None:
        if deps is None:
            deps = []

        self.tasks[name] = task
        self.deps[name] = deps

    def run(self, name: str, opts: Opts | None = None) -> None:
        opts = self.opts | opts
        logger.debug(f'Manager options: {opts}')

        exceptions = self._run(name, opts)
        if exceptions:
            raise RuntimeError(*exceptions)

    def _run(
        self,
        name: str,
        opts: Opts,
        already_run: set[str] | None = None,
        failed: set[str] | None = None,
        exceptions: set[Exception] | None = None,
    ) -> set[Exception]:
        already_run = already_run if already_run is not None else set()
        failed = failed if failed is not None else set()
        exceptions = exceptions if exceptions is not None else set()

        if name in already_run:
            logger.debug(f'Skipping deduped task: {name}')
            return exceptions
        already_run.add(name)
        logger.debug(f'Adding task to dedupe: {name}')

        deps = self.deps.get(name, [])
        if deps:
            logger.debug(f"Dependencies for '{name}': {' '.join(deps)}")
        for dep in deps:
            logger.debug(f"Running dependency for '{name}': {dep}")
            self._run(dep, opts, already_run, failed, exceptions)

        logger.debug(f'Current task: {name}')
        try:
            if name not in self.tasks:
                raise KeyError(f'No such task: {name}')

            failed_deps = [dep for dep in deps if dep in failed]
            if failed_deps:
                raise RuntimeError(
                    f"Dependencies failed for '{name}': "
                    f'{" ".join(failed_deps)}'
                )

            # NOTE: i don't know if we can even hit this condition
            # but i think we should check regardless
            # we can't run a task if the deps are not satisfied
            not_run_deps = [dep for dep in deps if dep not in already_run]
            if not_run_deps:
                raise RuntimeError(
                    f"Dependencies not run for '{name}': "
                    f'{" ".join(not_run_deps)}'
                )
            self.tasks[name].run(opts)

        # NOTE: i think all the error logging should be done here
        # at least for all errors that are supposed to crash
        # i.e. not ignore_err errors
        except KeyboardInterrupt as err:
            logger.error('Keyboard Interrupt')
            raise err
        except Exception as err:
            failed.add(name)
            exceptions.add(err)
            logger.error(err)
            if not opts.keep_going:
                raise err

        return exceptions



import abc
import tomllib
from pathlib import Path
from typing import Mapping

from . import Manager, Opts, SubprocessAction, Task


class Parser(abc.ABC):
    @abc.abstractmethod
    def parse(self, string: str) -> Manager: ...
    @abc.abstractmethod
    def validate_file(self, file: Path) -> bool: ...

    def find_config_file(self, root_path: Path) -> Path | None:
        for file in root_path.iterdir():
            if self.validate_file(file):
                return file
        return None


class PyprojectParser(Parser):
    def __init__(self, tool_name: str = 'faterunner') -> None:
        self.tool_name = tool_name

    # TODO: change assert to normal check and exceptions
    def parse(self, string: str) -> Manager:
        conf = tomllib.loads(string)

        assert 'tool' in conf
        assert self.tool_name in conf['tool']

        tool_config = conf['tool'][self.tool_name]
        assert isinstance(tool_config, Mapping)

        if 'options' in tool_config:
            assert isinstance(tool_config['options'], Mapping)
            opts = Opts(**tool_config['options'])
        else:
            opts = None
        manager = Manager(opts=opts)

        assert isinstance(tool_config['targets'], Mapping)
        for name, task_content in tool_config['targets'].items():
            if isinstance(task_content, list):
                actions = [SubprocessAction(cmd) for cmd in task_content]
                manager.add(name, Task(actions))
            elif isinstance(task_content, Mapping):
                commands = task_content.get('commands', [])
                opts = Opts(**task_content.get('options', {}))
                deps = task_content.get('dependencies', [])

                actions = [SubprocessAction(cmd) for cmd in commands]
                manager.add(name, Task(actions, opts))
                manager.deps[name] = deps
            else:
                raise TypeError(f'Target not list or mapping: {name}')

        return manager

    # validates if this is the file we want to choose
    # for example a pyproject.toml file can be our configuration file
    # if it has [tool.faterunner...] section, and if it doesn't we should
    # look elsewhere
    def validate_file(self, file: Path) -> bool:
        if file.name != 'pyproject.toml':
            return False
        string = file.read_text()
        conf = tomllib.loads(string)
        return 'tool' in conf and self.tool_name in conf['tool']



import argparse
import importlib.metadata
import logging
import sys
from pathlib import Path
from typing import NoReturn, Sequence

from . import Opts, parsers

_parsers: dict[str, parsers.Parser] = {
    **{
        entry_point.name: entry_point.load()()
        for entry_point in importlib.metadata.entry_points(
            group='faterunner.parsers'
        )
    },
    'pyproject': parsers.PyprojectParser(),
}


logger = logging.getLogger('faterunner.cli')


def setup_logging(level: str, verbosity: int = 0) -> None:
    # HACK: this setup is temporary

    if verbosity >= 1:
        format = '%(asctime)s %(name)s: [%(levelname)s] %(message)s'
    else:
        format = '%(name)s: [%(levelname)s] %(message)s'

    logging.basicConfig(format=format)
    logging.getLogger('faterunner').setLevel(level)


def guess_file_and_parser() -> tuple[Path, str]:
    for name, parser in _parsers.items():
        file = parser.find_config_file(Path.cwd())
        if file:
            return file, name
    raise RuntimeError("couldn't guess file and parser")


def guess_file(parser: str) -> Path:
    file = _parsers[parser].find_config_file(Path.cwd())
    if file:
        return file
    raise RuntimeError(f"couldn't guess the file for parser: {parser}")


def guess_parser(file: Path) -> str:
    for name, parser in _parsers.items():
        if parser.validate_file(file):
            return name
    raise RuntimeError(f"couldn't guess the parser for file: {file}")


def make_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description='the fate of a task is to be run'
    )

    opts_group = parser.add_argument_group()
    opts_group.add_argument(
        '--silent',
        action=argparse.BooleanOptionalAction,
        help='suppress output',
    )
    opts_group.add_argument(
        '--ignore-err',
        action=argparse.BooleanOptionalAction,
        help='ignore any error',
    )
    opts_group.add_argument(
        '--keep-going',
        action=argparse.BooleanOptionalAction,
        help='keep running tasks even if some cannot be done',
    )
    opts_group.add_argument(
        '--dry',
        action=argparse.BooleanOptionalAction,
        help='do not execute actions',
    )

    parser.add_argument(
        '--parser',
        choices=_parsers.keys(),
        help='parser to use (defaults to a guess)',
    )
    parser.add_argument(
        '-f',
        '--file',
        type=Path,
        help='configuration file (defaults to a guess)',
    )
    parser.add_argument(
        '-l', '--list', action='store_true', help='print all tasks and exit'
    )
    parser.add_argument(
        '--logging',
        default='INFO',
        choices=logging.getLevelNamesMapping().keys(),
        help='logging level (defaults to INFO)',
    )
    parser.add_argument(
        '-V',
        action='count',
        dest='verbosity',
        default=0,
        help='logging verbosity (more Vs more info)',
    )
    parser.add_argument('target', nargs='?', help='task to execute')
    return parser


def cli(argv: Sequence[str] | None = None) -> None:
    parser = make_parser()
    args = parser.parse_args(argv)
    setup_logging(args.logging, args.verbosity)

    try:
        if args.file is None and args.parser is None:
            args.file, args.parser = guess_file_and_parser()
            logger.debug(f'Guessed parser: {args.parser}')
            logger.debug(f'Guessed file: {args.file}')
        elif args.file is None:
            args.file = guess_file(args.parser)
            logger.debug(f'Guessed file: {args.file}')
        elif args.parser is None:
            args.parser = guess_parser(args.file)
            logger.debug(f'Guessed parser: {args.parser}')
    except RuntimeError as err:
        logger.critical(err.args[0])
        sys.exit(1)

    manager = _parsers[args.parser].parse(args.file.read_text())

    if args.list:
        print_and_exit('Targets:', *manager.tasks.keys(), exit_code=0)

    if args.target is None:
        args.target = tuple(manager.tasks.keys())[0]
        logger.debug(f'Guessed target: {args.target}')

    logger.debug(
        f'File: {args.file}, Parser: {args.parser}, Target: {args.target}',
    )

    opts = Opts(
        silent=args.silent,
        ignore_err=args.ignore_err,
        keep_going=args.keep_going,
        dry=args.dry,
    )
    logger.debug(f'Force options: {opts}')

    try:
        manager.run(args.target, opts)
    except (Exception, KeyboardInterrupt):
        sys.exit(1)


def print_and_exit(*values, exit_code: int = 1) -> NoReturn:
    file = sys.stdout if exit_code == 0 else sys.stderr
    print(*values, file=file)
    sys.exit(exit_code)


if __name__ == '__main__':
    cli(None)
